"""Functions for keeping track of collisions between particles
faster-than-naive-ly.

We've got bin_* algorithms for bin-based collision detection that I'll
implement first, and sort_* algorithms for an insertion-sort based collision
detection method that's fast if particles dont swap places very often."""
#TODO compare bin_* method with method where particle position array itself is
# binned in the original simulation code, for cache-help
import numpy as np
from numba import jit
import warnings
import unittest

def bin_init(particle_positions, max_particles_per_bin,
             num_bins_per_edge, bounds):
    """Returns a structure tracking which paritcles are in which bins.
    if max_particles_per_bin is violated, undefined behavior.
    bounds should be a tuple (len == ndims) of tuples (len == 2 ea) holding the
    min and max values for the coordinates of each dimension.
    num_bins_per_edge should be how many bins you'll have in each dimension."""
    ndims = len(bounds)
    num_particles, ndim_x = particle_positions.shape
    if ndims != ndim_x:
        raise ValueError("Number of dimensions of bounds provided {ndims:d} "
                         "does not match number of dimensions inferred from "
                         "coordinates of each particle position provided "
                         "{ndim_x:d}.".format(ndims=ndims, ndim_x=ndim+x))
    if ndims*num_bins_per_edge > num_particles:
        warnings.warn("More boxes than particles is inefficient! Use a "
                      "tree-based approach for finding which boxes are "
                      "currently in use if this is truly required.",
                      RuntimeWarning)
    # each bin corresponds to a list of integer indices to particles that are
    # in that bin and to a number of particles in that bin
    count_dims = [num_bins_per_edge for _ in range(ndims)]
    bin_dims = count_dims + [max_particles_per_bin]
    binned_particles = np.zeros(bin_dims, dtype=np.int)
    binned_counts = np.zeros(count_dims, dtype=np.int)
    # bin_walls = [np.linspace(min, max, num_bins_per_edge+1)
    #              for min, max in bounds]
    # transpose to help cacheing
    particle_bin_idxs = np.zeros((ndims, num_particles), dtype=np.int)
    xs = particle_positions.T
    for i, (min,max) in enumerate(bounds):
        for j, x in enumerate(xs):
            particle_bin_idxs[i,j] = binx(x, num_bins_per_edge, min, max)
    # untranspose to help caching
    particle_bin_idxs = particle_bin_idxs.T
    for i in range(num_particles):
        # since our box is a cube, with same number of discretizations on each
        # dimension, the multiidx->linear idx conversion is just a polynomial
        # evaluation, if we wanted to do it
        # bin_idx = np.polyval(particle_bin_idxs[i,:], num_bins_per_edge)
        c_idx = particle_bin_idxs[i,:]
        b_idx = c_idx + (binned_counts[c_idx],)
        binned_particles[b_idx] = i
        binned_counts[c_idx] += 1


def binx(x, nbins, min, max):
    """Our binning algorithm, separated out for the purposes of testing before
    vectorization."""
    return np.floor(nbins*(x - min)/(max - min)).astype(np.int)


class TestBinning(unittest.TestCase):
    """Suppose we have 4 bins, with boundaries 0,1,2,3,4.
    Then we know that the following x -> binx(x) should hold:
        -1 -> anything < 0
        0 -> 0
        0.5 -> 0
        1.0 -> 1
        4.0 -> anything > 3
    """
    test_min = 0
    test_max = 4
    nbins = 4
    bin_test = lambda x: binx(x, TestBinning.nbins, TestBinning.test_min, TestBinning.test_max)

    def test_binx_below_min(self):
        self.assertLess(TestBinning.bin_test(-1), 0)

    def test_binx_equal_min(self):
        self.assertEqual(TestBinning.bin_test(0), 0)

    def test_binx_equal_internal_bdry(self):
        self.assertEqual(TestBinning.bin_test(1), 1)

    def test_binx_equal_max(self):
        self.assertGreater(TestBinning.bin_test(4), 3)



r"""
Simulation twistable, stretchable-shearable wormlike chains.

Notes
-----
The ssWLC is implemented as described in `Koslover et. al., Soft Matter, 2013,
9, 7016
<https://pdfs.semanticscholar.org/87e5/64cd9b84db3ddf38a869c14de09fa2e05de2.pdf>`_.
This defines the locations (:math:`r^{(i)}(t_j)`) and tangent vectors
(:math:`u^{(i)}(t_j)`) of each (:math:`i`\th) bead at each time
point (:math:`t_j`).

The twist energy is (loosely speaking) quadratic in the angle that one must
rotate the material normals of one bead to get them to align with the next bead
(after the two beads have been rotated once to make the tangent vectors align).

There are other strategies one might take for this, but we will leave them as
"#TODO" in the code for now.
"""
from ..tabulation import dsswlc_from_del

# #TODO Alternate twist versions
# 1) Based on Hermite polynomials.
# The other two material normals, (the "twistable" part of the chain) are
# implemented independently. Since the ssWLC specifies the locations and
# tangent vectors at each bead, the chain is approximated by a cubic Hermite
# polynomial between each pair of beads. This cubic Hermite polynomial has a
# natural (Levi-Civita) connection relating the material normals at each bead
# to a pair of material normals at the next bead corresponding to "zero twist".
# The twist energy of each segment is quadratic in the angle about the tangent
# vector (:math:`\theta = \int \omega_3 ds`) by which the curve deviates from
# this natural frame.
# 2) Include bend/twist and shear/twist coupling.

# # Fortran parameters from old code
# EB=PARA(1)
# EPAR=PARA(2)
# EPERP=PARA(3)
# GAM=PARA(4)
# ETA=PARA(5)
# XIR=PARA(6)
# XIU=PARA(7)
# LBOX=PARA(8)
# LHC=PARA(9)
# VHC=PARA(10)
# MAGR=sqrt(XIR*2.0/DT)
# MAGU=sqrt(XIU*2.0/DT)
# DRPAR = (R[i+1] - R[i]).U[i]
# DRPERP = (R[i+1] - R[i]) - DRPAR*U[i]
# FRAND = MAGR*rnorm()
# TRAND = MAGU*rnorm(), projected to be perpendicular to u
# GR = (U[i+1] - (U[i+1].U[i])U[i] - ETA*DRPERP
# F = -ETA*EB*GR _ EPAR*(DRPAR - GAM)*U[i] + EPERP*DRPERP
# FELAS[i] += F, FELAS[i+1] -= F
# GU = U[i+1] - U[i] - ETA*DRPERP
# T1 = EB*GU
# T2 = -ETA*EB*DRPAR*DU + ETA*EB*(1 - U[i+1].U[i])*DR
#      - EPAR*(DRPAR - GAM)*DR + EPERP*DRPAR*DRPERP
# TELAS[i] += T1 + T2, TELAS[i+1] -= T1
# actual force/torque is /= XI[U/R]

def sswlc(N, L, lp, t, t_save):
    """
    Simulate a stretchable-shearable WLC.

    Parameters
    ----------
    N : int
        The number of beads to use.
    L : float
        The length of the chain.
    lp : float
        The persistence length of the underlying WLC being approximated. (Must
        be same units as *L*.
    t : array_like
        The times for which to simulate the polymer's motion.
    t_save : (M, ) array_like
        The subset of *t* for which we should save the output.

    Returns
    -------
    r : (M, N, 3) array of float
        The positions of each bead at each save time.
    u : (M, N, 3) array of float
        The tangent vectors to the underlying wormlike chain at each bead.
    """
    L0 = L / (N - 1)
    delta = L0/lp
    e_b, gam, e_par, e_perp, eta, xi_u, _ = dsswlc_from_del(delta)
    # same normalization done in the fortran code for now
    e_b = lp*e_b/(delta*lp)
    e_par = e_par/(delta*lp*lp)
    e_perp = e_perp/(delta*lp*lp)
    gam = delta*lp*gam
    eta = eta/lp
    xi_u = xi_u*delta
    xi_r = delta
    # Lena's estimate of minimum dt possible
    max_dt = (1/2)*xi_u/(e_perp*gam**2)
    # e_twist = lt/(delta/lp)
    if np.any(np.diff(t) > max_dt):
        raise ValueError(f"Maximum recommended time step is {max_dt}")
    return _jit_sswlc_lena(N, L, lp, t, t_save, e_b, gam, e_par, e_perp,
                           eta, xi_u, xi_r)


# @jit(nopython=True)
def _jit_sswlc_lena(N, L, lp, t, t_save, e_b, gam, e_par, e_perp,
                    eta, xi_u, xi_r):
    """
    TODO: test
    """
    rtol = 1e-5
    # initial position
    r0, u0 = _init_circle(N, L)
    # pre-alloc output
    if t_save is None:
        t_save = t
    r = np.zeros(t_save.shape + r0.shape)
    u = np.zeros(t_save.shape + r0.shape)
    # setup for saving only requested time points
    save_i = 0
    if t[0] == t_save[save_i]:
        r[0] = r0
        u[0] = u0
        save_i += 1
    # at each step i, we use data (r,t)[i-1] to create (r,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old r (r[i-1]) "r0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = t[i] - t[i-1]
        F_brown = np.sqrt(2*xi_r/h)*np.random.randn(*r0.shape)
        T_brown = np.sqrt(2*xi_u/h)*np.random.randn(*u0.shape)

        F_elas, T_elas = ft_elas_sswlc(r0, u0, e_b, gam, e_par, e_perp, eta,
                                       xi_u, xi_r)
        dr1 = (F_elas + F_brown)/xi_r  # slope at beginning of time step
        du1 = (T_elas + T_brown)/xi_u
        for j in range(N):
            du1[j] -= (du1[j]@u0[j])*u0[j]
        r_est = r0 + dr1*(h/2)  # first estimate at midpoint
        u_est = u0 + du1*(h/2)
        u_est = u_est/np.linalg.norm(u_est, axis=1)[:, None]

        F_elas, T_elas = ft_elas_sswlc(r_est, u_est, e_b, gam, e_par, e_perp,
                                       eta, xi_u, xi_r)
        dr2 = (F_elas + F_brown)/xi_r  # first estimate slope at midpoint
        du2 = (T_elas + T_brown)/xi_u
        for j in range(N):
            du2[j] -= (du2[j]@u_est[j])*u_est[j]
        r_est = r0 + dr2*(h/2)  # second estimate at midpoint
        u_est = u0 + du2*(h/2)
        u_est = u_est/np.linalg.norm(u_est, axis=1)[:, None]

        F_elas, T_elas = ft_elas_sswlc(r_est, u_est, e_b, gam, e_par, e_perp,
                                       eta, xi_u, xi_r)
        dr3 = (F_elas + F_brown)/xi_r  # second estimate slope at midpoint
        du3 = (T_elas + T_brown)/xi_u
        for j in range(N):
            du3[j] -= (du3[j]@u_est[j])*u_est[j]
        r_est = r0 + dr3*h  # estimate at endpoint
        u_est = u0 + du3*h
        u_est = u_est/np.linalg.norm(u_est, axis=1)[:, None]

        F_elas, T_elas = ft_elas_sswlc(r_est, u_est, e_b, gam, e_par, e_perp,
                                       eta, xi_u, xi_r)
        dr4 = (F_elas + F_brown)/xi_r  # estimate slope at next time point
        du4 = (T_elas + T_brown)/xi_u
        for j in range(N):
            du4[j] -= (du4[j]@u_est[j])*u_est[j]

        # average the slope estimates
        r0 = r0 + h * (dr1 + 2*dr2 + 2*dr3 + dr4)/6
        u0 = u0 + h * (du1 + 2*du2 + 2*du3 + du4)/6
        u0 = u0/np.linalg.norm(u0, axis=1)[:, None]

        # save if at a time in t_save
        if np.abs(t[i] - t_save[save_i]) < rtol*np.abs(t_save[save_i]):
            r[save_i] = r0
            u[save_i] = u0
            save_i += 1
            if save_i >= len(t_save):
                break
    return r, u


# @jit(nopython=True)
def _jit_sswlc_clean(N, L, lp, t, t_save, e_b, gam, e_par, e_perp,
                     eta, xi_u, xi_r):
    """
    WARNING: known to be buggy...
    """
    rtol = 1e-5
    # initial position
    r0 = np.zeros((N, 3))
    r0[:, 0] = np.linspace(0, L, N)
    u0 = np.zeros((N, 3))
    u0[:, 0] = 1
    # pre-alloc output
    if t_save is None:
        t_save = t
    r = np.zeros(t_save.shape + r0.shape)
    u = np.zeros(t_save.shape + r0.shape)
    # setup for saving only requested time points
    save_i = 0
    if t[0] == t_save[save_i]:
        r[0] = r0
        u[0] = u0
        save_i += 1
    # at each step i, we use data (r,t)[i-1] to create (r,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old r (r[i-1]) "r0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = t[i] - t[i-1]
        F_elas, T_elas = ft_elas_sswlc(r0, u0, e_b, gam, e_par, e_perp, eta,
                                       xi_u, xi_r)
        dW_r = np.random.randn(*r0.shape)
        dW_u = np.random.randn(*u0.shape)
        # -1 or 1, p=1/2
        S_r = 2*(np.random.rand() < 0.5) - 1
        S_u = 2*(np.random.rand() < 0.5) - 1
        # D = sigma^2/2 ==> sigma = np.sqrt(2*D)
        F_brown = np.sqrt(2*xi_r/h)*(dW_r - S_r)
        T_brown = np.sqrt(2*xi_u/h)*(dW_u - S_u)
        for j in range(N):
            T_brown[j] -= (T_brown[j]@u0[j])*u0[j]
        # estimate for slope at interval start
        K1_r = (F_elas + F_brown)/xi_r
        K1_u = (T_elas + T_brown)/xi_u
        r1 = r0 + h*K1_r
        u1 = u0 + h*K1_u
        for j in range(N):
            u1[j] = u1[j]/np.linalg.norm(u1[j])
        # estimate for slope at interval end
        F_brown = np.sqrt(2*xi_r/h)*(dW_r + S_r)
        T_brown = np.sqrt(2*xi_u/h)*(dW_u + S_u)
        for j in range(N):
            T_brown[j] -= (T_brown[j]@u0[j])*u0[j]
        F_elas, T_elas = ft_elas_sswlc(r1, u1, e_b, gam, e_par, e_perp, eta,
                                       xi_u, xi_r)
        K2_r = (F_elas + F_brown)/xi_r
        K2_u = (T_elas + T_brown)/xi_u
        # average the slope estimates
        r0 = r0 + h * (K1_r + K2_r)/2
        u0 = u0 + h * (K1_u + K2_u)/2
        for j in range(N):
            u0[j] = u0[j]/np.linalg.norm(u0[j])
        if np.abs(t[i] - t_save[save_i]) < rtol*np.abs(t_save[save_i]):
            r[save_i] = r0
            u[save_i] = u0
            save_i += 1
            if save_i >= len(t_save):
                break
    return r, u



"""Combining collision times statistics from CSV's generated by wlcsim.Scan"""
from . import data as wdata
from .utils.multiprocessing_mapreduce import SimpleMapReduce
import logging
import os
import pandas as pd
import glob
import numpy as np
import itertools

logger = logging.getLogger(__name__)

default_group_params_coltimes = ['linear_distance']
def map_mean_coltimes(sim_dir, group_params=default_group_params_coltimes):
    sim = wdata.Sim(sim_dir)
    try:
        for param in group_params:
            if param not in sim.coltimes.columns:
                sim.coltimes[param] = sim.params[param]
        gb_coltimes = sim.coltimes.groupby(group_params, as_index=True).coltime
    except OSError:
        # no coltimes in this simulation dir
        return []
    means = gb_coltimes.mean()
    nan_counts = gb_coltimes.agg(lambda x: np.sum(np.isnan(x)))
    total_counts = gb_coltimes.size()
    df = pd.concat([means, nan_counts, total_counts], axis=1,
                   keys=['mean_coltime', 'coltime_nan_count', 'coltime_count'])
    # this will flatten a multi-index if it exists, else no-op
    df.index = df.index.values
    # with a flattened multiindex, we can transpose to get out the three values
    # as a Series
    return list(df.T.items())

class MapMeanColtimes:
    """Function object, allows the same-named method to be passed to
    Pool.map"""
    def __init__(self, group_params):
        self.group_params = group_params
    def __call__(self, sim_dir):
        return map_mean_coltimes(sim_dir, self.group_params)


def reduce_mean_coltimes(item):
    param_vals, coltime_stats = item
    # concat will put the param_vals as the column (then .T to row) names
    df = pd.concat(coltime_stats, axis=1).T
    # all indexes will match (== param_vals), so we will get one row at end
    row = df.groupby(df.index).agg({'mean_coltime': np.mean,
                                    'coltime_nan_count': np.sum,
                                    'coltime_count': np.sum})
    # the key of the single row is param_vals, so we can just
    # return next(row.T.items())
    # but honestly, keeping it in tuple form is annoying, since we will
    # just concatenate all the reduced results afterwards anyway, so we just
    return row

def reduce_rigid_mean_coltimes(item):
    param_vals, coltime_stats = item
    # concat will put the param_vals as the column (then .T to row) names
    df = pd.concat(coltime_stats, axis=1).T
    # all indexes will match (== param_vals), so we will get one row at end
    df = df[df['nan_counts'] == 0]
    row = df.groupby(df.index).agg({'mean_coltime': np.mean,
                                    'coltime_nan_count': np.sum,
                                    'coltime_count': np.sum})
    # the key of the single row is param_vals, so we can just
    # return next(row.T.items())
    # but honestly, keeping it in tuple form is annoying, since we will
    # just concatenate all the reduced results afterwards anyway, so we just
    return row

def striding_map_mean_coltimes(sim_dir, idx_gap=0,
                               group_params=['linear_distance']):
    """Not yet implemented. In the future, this should calculate the mean
    for a particular simulation of all the beads of each separation possible,
    without overlapping looped sections. Gah. Don't feel like typing that
    better."""
    pass

def map_reduce_mean_coltimes(run_dir, num_workers=None,
                             group_params=default_group_params_coltimes,
                             be_rigid=False, pool_type='process'):
    # write necessary csv files if not already there
    wdata.write_coltimes_csvs(run_dir)
    # get list of simulation directories to map over
    sim_dirs = wdata.Scan(run_dir).sim_paths
    # construct the MapReduce-r
    map_func = MapMeanColtimes(group_params)
    if be_rigid:
        reduce_func = reduce_rigid_mean_coltimes
    else:
        reduce_func = reduce_mean_coltimes
    mapper = SimpleMapReduce(map_func, reduce_func, num_workers=num_workers)
    # return a dataframe containing all the stats calculated
    df = pd.concat(mapper(sim_dirs), axis=0)
    df.index = pd.MultiIndex.from_tuples(df.index, names=group_params)
    return df


def map_cdf_coltimes(sim_dir, dt, group_params=default_group_params_coltimes):
    """TODO: actually implement"""
    sim = wdata.Sim(sim_dir)
    try:
        for param in group_params:
            if param not in sim.coltimes.columns:
                sim.coltimes[param] = sim.params[param]
        gb_coltimes = sim.coltimes.groupby(group_params, as_index=True).coltime
    except OSError:
        # no coltimes in this simulation dir
        return []

    df = pd.concat([], axis=1,
                   keys=['mean_coltime', 'coltime_nan_count', 'coltime_count'])
    # this will flatten a multi-index if it exists, else no-op
    df.index = df.index.values
    return list(df.T.items())

# non map-reduced version. Get's pretty bogged down by disk speed already
# based on watching it in htop.
def get_com_msd(run_dir):
    sim_dirs = wdata.Scan(run_dir).sim_paths
    num_sims = len(sim_dirs)
    sim0 = wdata.Sim(sim_dirs[0])
    num_time_points, ndims, num_polymers, num_beads = sim0.r.shape
    msd_com = np.zeros((num_time_points, ndims, num_polymers))
    for i,sim_dir in enumerate(sim_dirs):
        sim = wdata.Sim(sim_dir)
        msd_com += np.power(sim.center_of_mass - sim.center_of_mass[0,:,:], 2)
        print('Progress: {:2.1%}'.format(float(i)/num_sims), end='\r')
    msd_com = msd_com/len(sim_dirs)
    return msd_com

def fraction_collided_from_csv(csv_file):
    with open(csv_file) as f:
        # first establish the correct column identities
        column_names = next(f).rstrip().split(',')
        col_i = None
        i_i = None
        j_i = None
        nb_i = None
        for i,name in enumerate(column_names):
            if name.upper() == 'COLTIME':
                col_i = i
            elif name.upper() == 'I':
                i_i = i
            elif name.upper() == 'J':
                j_i = i
            elif name.upper() == 'NB' or name.upper() == 'N':
                nb_i = i
        if col_i is None or i_i is None or j_i is None or nb_i is None:
            raise ValueError(csv_file + "does not contain coltime,i,j,nb columns.")
        # peek at first line to get number of beads per polymer
        try:
            data = next(f).rstrip().split(',')
        except StopIteration:
            raise ValueError(csv_file + "has no data")
        nB = int(data[nb_i])
        num_uncollided = np.zeros((nB))
        num_precollided = np.zeros((nB))
        num_total = np.zeros((nB))
        # loop and a half, since we peeked the first line
        coltime = float(data[col_i])
        i = int(round(float(data[i_i])))
        j = int(round(float(data[j_i])))
        linear_distance = i - j
        num_total[linear_distance] += 1
        if coltime == 0.0:
            num_precollided[linear_distance] += 1
        elif coltime < 0.0:
            num_uncollided[linear_distance] += 1
        for line in f:
            data = line.rstrip().split(',')
            coltime = float(data[col_i])
            i = int(round(float(data[i_i])))
            j = int(round(float(data[j_i])))
            linear_distance = i - j
            num_total[linear_distance] += 1
            if coltime == 0.0:
                num_precollided[linear_distance] += 1
            elif coltime < 0.0:
                num_uncollided[linear_distance] += 1
    return num_uncollided, num_precollided, num_total

def get_coltimes_by_pos(scan_dir, linear_distance):
    """Get all coltimes that happened at a fixed linear distance as a
    function of their position on the polymer"""
    scan = wdata.Scan(scan_dir)
    c_fixed_dist = pd.DataFrame(columns=['coltime', 'i', 'j',
                                'linear_distance'])
    for sim_dir in scan.sim_dirs:
        s = wdata.Sim(os.path.join(scan.scan_dir, sim_dir))
        try:
            c = s.coltimes
        except FileNotFoundError:
            continue
        c_fixed_dist = c_fixed_dist.append(
            c.loc[c['linear_distance'] == linear_distance]
        )
    return c_fixed_dist



import sys
import re

from setuptools import setup, find_packages
import versioneer

long_description = open("README.rst").read()
long_description = re.sub(':ref:', '', long_description)
long_description = re.sub(':mod:', '', long_description)

if __name__ == "__main__":
    if sys.version_info < (3, 5):
        raise ImportError("requires Python>=3.5")

    setup(name="wlcsim",
          author="Bruno Beltran",
          author_email="brunobeltran0@gmail.com",
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass(),
          url="https://github.com/SpakowitzLab/BasicWLC",
          license="",
          long_description=long_description,
          long_description_content_type='text/x-rst',
          classifiers=["Intended Audience :: Science/Research",
                       'Intended Audience :: Developers',
                       'Development Status :: 2 - Pre-Alpha',
                       'License :: OSI Approved :: MIT License',
                       "Natural Language :: English",
                       "Programming Language :: Python :: 3 :: Only",
                       "Programming Language :: Fortran",
                       "Topic :: Scientific/Engineering :: Chemistry",
                       "Topic :: Scientific/Engineering :: Physics",
                       ],
          packages=find_packages(include=["wlcsim", "wlcsim.*"]),
          package_data={
              'wlcsim.tabulation': ['*.csv', 'dssWLCparams'],
          },
          install_requires=["scipy", "statsmodels", "matplotlib", "seaborn",
                            "mpmath", "pandas", "numpy", "bruno_util",
                            "spycial", "PyQt5"],
          )



r"""
Simulate Rouse polymers.

Notes
-----
There are various ways to parameterize Rouse polymers. In this module, we use
the convention that `N` is the number of beads of a Rouse polymer (contrary to
e.g. Doi & Edwards, where `N` is the number of Kuhn lengths in the polymer). We
instead use `Nhat` for the number of Kuhn lengths in the polymer.

.. warning:

    This is different from the convention in :mod:`wlcsim.analytical.rouse`,
    where :math:`N` is the number of Kuhn lengths in the polymer, to match the
    usual conventions in e.g. Doi & Edwards.

`D` is the "diffusivity of a Kuhn length", i.e. `kbT/xi`, where `xi` is the
dynamic viscosity of the medium in question, as typically found in the Langevin
equation for a Rouse polymer

    .. math::

        \xi \frac{d}{dt} \vec{r}(n, t)
        = k \frac{d^2}{dn^2} \vec{r}(n, t) + f^{(B)}

This means that to simulate a Rouse polymer diffusing in a medium with
viscosity `xi`, the diffusion coefficient of each bead should be set to `Dhat =
D (N/Nhat)`. Since :math:`\xi` is in units of "viscosity per Kuhn length" and
`(N/Nhat)` is in units of "number of beads over number of Kuhn lengths", this
can be thought of as changing units from "viscosity per Kuhn length" to
"viscosity per bead".

Some books use `b` for the mean (squared) bond length between beads in the
discrete gaussian chain. We instead use `b` to be the real Kuhn length of the
polymer, so that the mean squared bond length `bhat**2` is instead given by
`L0*b`, where `L0` is L0 = L/(N-1) is the amount of polymer "length"
represented by the space between two beads.

The units chosen here make the most sense if you don't actually consider the
polymer to be a "real" Rouse polymer, but instead an e.g. semiflexible chain
with a real "length" whose long distance statistics are being captured by the
simulation.

To compare these results to the :mod:`rouse.analytical.rouse` module, use

>>> plt.plot(t_save, sim_msd)
>>> plt.plot(t_save, wlcsim.analytical.rouse.rouse_mid_msd(t_save, b, Nhat, D,
...                                                        num_modes=int(N/2)))
>>> plt.plot(t_save, 6*Dhat*t_save)
>>> plt.plot(t_save, 6*(Dhat/N)*t_save)  # or 6*(D/Nhat)*t_save
>>> # constant prefactor determined empirically...
>>> # no, using the value of 6*(Dhat/N)*t_R doesn't work...
>>> plt.plot(t_save, 1.9544100*b*np.sqrt(D)*np.sqrt(t_save))

where cutting off the number of modes corresponds to ensuring that the rouse
behavior only continues down to the length scale of a single "bead", thus
matching the simulation at arbitrarily short time scales. (Notice that we abide
by the warning above. Our `Nhat` is :mod:`wlcsim.analytical.rouse`'s `N`).

Example
-------

For example, if you want to simulate a megabase of DNA, which has a real linear
length of about 1e6 megabase/base * 0.33 nm/base ~ 3e5 nm, and a Kuhn length of
1e2 nm, then

"""
from .runge_kutta import rk4_thermal_lena
from .init_beads import init_linear_rouse_conf
from .forces import f_conf_ellipse, f_elas_linear_rouse

from numba import njit
import numpy as np


def recommended_dt(N, L, b, D):
    r"""
    Recommended "dt" for use with rouse*jit family of functions.

    Currently set to :math:`\frac{1}{10}\frac{b^2}{6D}`.

    See the :func:`jit_srk1` docstring for source of this time scale.
    """
    Nhat = L/b
    L0 = L/(N-1)
    Dhat = D*N/Nhat
    bhat = np.sqrt(L0*b)
    return (1/10)*bhat**2/(6*Dhat)


def terminal_relaxation(N, L, b, D):
    Nhat = L/b
    return b**2 * Nhat**2 / D


def measured_D_to_rouse(Dapp, d, N, bhat=None, regime='rouse'):
    r"""
    Get the full-polymer diffusion coefficient from the "apparent" D.

    In general, a discrete Rouse polymer's MSD will have three regimes. On time
    scales long enough that the whole polymer diffuses as a large effective
    particle, the MSD is of the form :math:`6 D/\hat{N} t`, where, as is true
    throughout this module, :math:`D` is the diffusivity of a Kuhn length, and
    :math:`\hat{N}` is the number of Kuhn lengths in the polymer.

    This is true down to the full chain's relaxation time (the relaxation time
    of the 0th Rouse mode, also known as the "Rouse time") :math:`t_R = N^2 b^2
    / D`. For times shorter than this, the MSD will scale as :math:`t^{1/2}`.
    Imposing continuity of the MSD, this means that the MSD will behave as
    :math:`\kappa_0 D/\hat{N} (t_R t)^{1/2}`, where :math:`\kappa_0` is a
    constant that I'm too lazy to compute using :math:`\lim_{t\to 0}` of the
    analytical Rouse MSD result, so I just determine it empirically. We rewrite
    this MSD as :math:`\kappa b D^{-1/2} t^{1/2}`, and find by comparing to the
    analytical Rouse theory that :math:`\kappa = 1.9544100(4)`.

    Eventually (at extremely short times), most "real" polymers will eventually
    revert to a MSD that scales as :math:`t^1`. This cross-over time/length
    scale defines a "number of segments" :math:`\tilde{N}`, where the
    diffusivity of a segment of length :math:`\tilde{L} = L/(\tilde{N} - 1)`
    matches the time it takes stress to propagate a distance :math:`\tilde{L}`
    along the polymer. Said in other words, for polymer lengths smaller than
    :math:`\tilde{L}`, the diffusivity of the segment outruns the stress
    communication time between two neighboring segments.

    The diffusivity at these extremely short times will look like :math:`6 D
    (\tilde{N} / \hat{N}) t`. In order to simulate this behavior exactly, one
    can simply use a polymer with :math:`\tilde{N}` beads, then all three
    length scales of the MSD behavior will match.

    This three-regime MSD behavior can be very easily visualized in log-log
    space, where it is simply the continuous function defined by the following
    three lines. From shortest to longest time scales, :math:`\log{t}
    + \log{6D(\tilde{N}/\hat{N})}`, :math:`(1/2)\log{t} + \log{\kappa b
    D^{1/2}}`, and :math:`\log{t} + \log{6D\hat{N}}`. The lines
    (in log-log space), have slopes 1, 1/2, and 1, respectively, and their
    "offsets" (y-intercept terms with the log removed) are typically referred
    to as :math:`D_\text{app}`.

    The simulations in this module use the diffusivity of a single Kuhn length
    as input (i.e. plain old :math:`D` is expected as input) but typically,
    measurements of diffusing polymers are done below the Rouse time, and for
    long enough polymers (like a mammalian chromosome), it may be difficult to
    impossible to measure unconfined diffusion at time scales beyond the Rouse
    time. This means you often can't just look at the diffusivity of the whole
    polymer and multiply by :math:`\hat{N}` to get the diffusivity of a Kuhn
    length. And since there's not really a principled way in general to guess
    :math:`\tilde{N}`, you usually can't use the short-time behavior to get
    :math:`D` either, even supposing you manage to measure short enough times
    to see the cross-over back to :math:`t^1` scaling. This means that usually
    the only way one can extract :math:`D` is by measuring
    :math:`D_\text{app}`, since we can use the fact that
    :math:`D_\text{app} = \kappa b D^{1/2}` (the :math:`\kappa` value
    quoted above only works for 3-d motion. I haven't bothered to check if it
    scales linearly with the number of dimensions or like :math:`\sqrt{d}` for
    :math:`d`-dimensional motion, but this shouldn't be too hard if you are
    interested).

    In short, this function gives you :math:`D` given :math:`D_\text{app}`.

    Parameters
    ----------
    D_app : float
        The measured :math:`D_\text{app} based on the y-intercept of the MSD in
        log-log space. This can be computed as `np.exp(scipy.stats.linregress(
        np.log(t), np.log(msd))[1])`, assuming that `t` is purely in the regime
        where the MSD scales like :math:`t^{1/2}`.
    Nhat : int
        number of beads in the polymer
    d : int
        number of dimensions of the diffusion (e.g. 3 in 3-d space)

    Returns
    -------
    D_rouse : float
        the diffusivity of a Kuhn length of the polymer

    Notes
    -----
    What follows is an alternate way of "thinking" about the three regimes of
    the MSD, from a simulation-centric perspective, that I typed up back when I
    first wrote this function, but has been obsoleted by the description at the
    start of this docstring.

    In general, a discrete Rouse polymer's MSD will have three regimes. On time
    scales short enough that chain connectivity is not relevant, the MSD is of
    the form :math:`6 \hat{D} t`.

    As soon as chain connectivity begins to affect the dynamics, the form of
    the MSD will change to :math:`\kappa \hat{b} \hat{D}^{1/2} t^{1/2}`, where
    we have determined the constant :math:`\kappa` empirically (using our exact
    analytical theory) to be approximately 1.9544100(4). Note that (up to a
    constant factor), this is just :math:`6 \hat{D} (t_R t)^{1/2}`, where
    :math:`t_R` is the relaxation time (aka Rouse time) of the polymer, given
    by :math:`\hat{b}^2 \tilde{N}^2 / \hat{D}`.

    This is because the Rouse behavior only continues up to the relaxation time
    :math:`t = t_R`, where the MSD transitions into the form
    :math:`6 \frac{\hat{D}}{\tilde{N}} t`. If you ask what "line" with slope
    1/2 in log-log space intersects :math:`6 \frac{\hat{D}}{\tilde{N}} t` at
    :math:`t = t_R`, you get the above form for the sub-Rouse time MSD.

    Here, :math:`\tilde{N}` is distinguished from :math:`N` only as a
    formality. For a simulation of a discrete Rouse polymer with a fixed number
    of beads, :math:`N = \tilde{N}` is exactly the number of beads. For a
    "real" polymer, it is simply a numerical parameter that describes how long
    the Rouse behavior lasts (in the limiit of short times). For a "true" Rouse
    polymer (the fractal object) N is infinity and there is no time scale on
    which the MSD transitions back into :math:`t^1` behavior. But, for any real
    polymer, this will "almost always" eventually happen.
    """
    if d != 3:
        raise ValueError("We've only calculated kappa for d=3")
    kappa = 1.9544100
    return (Dapp/(kappa*bhat))**2


def with_integrator(N, L, b, D, t, x0=None, integrator=rk4_thermal_lena):
    r"""
    Simulate a Rouse polymer made of N beads free in solution.

    Parameters
    ----------
    N : float
        Number of beads in the chain.
    L : float
        Length of chain.
    b : float
        Kuhn length of the chain (same units as *L*).
    D : float
        Diffusion coefficient. (Units of ``length**2/time``). In order to
        compute *D* from a single-locus MSD, use `measured_D_to_rouse`.
    t : (Nt,) array_like of float
        Time points to use for stepping the integrator. None of our current
        integrators is currently capable of taking time steps larger than the
        time scale of the highest Rouse mode of the finite chain (use
        `~.bd.recommended_dt` to compute the optimal *dt*).
    x0 : (N, 3) array_like of float, optional
        The initial locations of each bead. If not specified, defaults to
        initializing from the free-draining equilibrium, with the first bead at
        the origin.
    integrator : Callable[[ForceFunc, float, Times, Positions], Positions]
        Either `~.runge_kutta.rk4_thermal_lena` or `~.runge_kutta.srk1_roberts`

    Returns
    -------
    (Nt, N, 3) array_like of float
        The positions of the *N* beads at each of the *Nt* time points.

    Notes
    -----
    The polymer beads can be described by the discrete Rouse equations

    .. math::
        \xi \frac{dx(i, t)}{dt} = - k (x(i, t) - x(i+1, t))
                                  - k (x(i, t) - x(i-1, t))
                                  + R(t)

    where :math:`\xi = k_BT/D`, :math:`k = 3k_BT/b^2`, :math:`b` is the Kuhn
    length of the polymer, :math:`D` is the self-diffusion coefficient of a
    bead, and :math:`R(t)/\xi` is a delta-correlated stationary Gaussian
    process with mean zero and :math:`\langle R(t) R(t')/\xi^2 \rangle =
    2DI\delta(t-t')`.

    Notice that in practice, :math:`k/\xi = 3D/b^2`, so we do not need to
    include mass units (i.e. there's no dependence on :math:`k_BT`).
    """
    # derived parameters
    L0 = L/(N-1)  # length per bead
    bhat = np.sqrt(L0*b)  # mean squared bond length of discrete gaussian chain
    Nhat = L/b  # number of Kuhn lengths in chain
    Dhat = D*N/Nhat  # diffusion coef of a discrete gaussian chain bead
    k_over_xi = 3*Dhat/bhat**2
    if x0 is None:
        x0 = bhat/np.sqrt(3)*np.cumsum(np.random.normal(size=(N, 3)), axis=0)
    # zero = np.array([[0, 0, 0]])

    @njit
    def rouse_f(x, t):
        N, _ = x.shape
        f = np.zeros(x.shape)
        for j in range(1, N):
            for n in range(3):
                f[j, n] += -k_over_xi*(x[j, n] - x[j-1, n])
                f[j-1, n] += -k_over_xi*(x[j-1, n] - x[j, n])
        return f
        # dx = np.diff(x, axis=0)
        # return -k_over_xi*(
        #     np.concatenate([zero, dx]) + np.concatenate([dx, zero])
        # )

    X = integrator(rouse_f, Dhat, t, x0)
    return X


@njit
def jit_srk1(N, L, b, D, t, t_save=None):
    r"""
    Faster version of `wlcsim.bd.rouse.rouse` using jit.

    ``N=101,L=100,b=1,D=1`` takes about 3.5min to run when
    ``t=np.linspace(0, 1e5, 1e7+1)``
    adding ``t_save`` does not slow function down

    Our srk1 scheme is accurate as long as :math:`\Delta t` is less than the
    transition time where the MSD goes from high k behavior (:math:`t^1`) to
    Rouse behavior (:math:`t^{1/2}`).  This is exactly the time required to
    diffuse a Kuhn length, so we just need
    :math:`\Delta t < \frac{\hat{b}^2}{6\hat{D}}` in 3D. the "crossover" from
    fast-k to rouse-like behavior takes about one order of magnitude in time,
    but adding more than that doesn't seem to make the MSD any more accurate,
    so we suggest setting :math:`\Delta t` to one tenth of the bound above.

    the number of orders of magnitude of "rouse" scaling the simulation will
    capture is exactly dictated by the ratio between this time scale and the
    rouse relaxation time (so like ``N**2``?)

    recall (doi & edwards, eq 4.25) that the first mode's relaxation time is
    :math:`\tau_1 = \frac{\xi N^2}{k \pi^2 }`.
    and the :math:`p`\th mode is :math:`\tau_p = \tau_1/p^2` (this is the
    exponential falloff rate of the :math:`p`\th mode's correlation function).
    """
    rtol = 1e-5
    # derived parameters
    L0 = L/(N-1)  # length per bead
    bhat = np.sqrt(L0*b)  # mean squared bond length of discrete gaussian chain
    Nhat = L/b  # number of Kuhn lengths in chain
    Dhat = D*N/Nhat  # diffusion coef of a discrete gaussian chain bead
    k_over_xi = 3*Dhat/bhat**2
    # initial position, sqrt(3) since generating per-coordinate
    x0 = bhat/np.sqrt(3)*np.random.randn(N, 3)
    # for jit, we unroll ``x0 = np.cumsum(x0, axis=0)``
    for i in range(1, N):
        x0[i] = x0[i-1] + x0[i]
    if t_save is None:
        t_save = t
    x = np.zeros(t_save.shape + x0.shape)
    dts = np.diff(t)
    # -1 or 1, p=1/2
    S = 2*(np.random.rand(len(t)) < 0.5) - 1
    save_i = 0
    if 0 == t_save[save_i]:
        x[0] = x0
        save_i += 1
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = dts[i-1]
        dW = np.random.randn(*x0.shape)
        # D = sigma^2/2 ==> sigma = np.sqrt(2*D)
        Fbrown = np.sqrt(2*Dhat/h)*(dW - S[i])
        # estimate for slope at interval start
        f = np.zeros(x0.shape)
        for j in range(1, N):
            for n in range(3):
                f[j, n] += -k_over_xi*(x0[j, n] - x0[j-1, n])
                f[j-1, n] += -k_over_xi*(x0[j-1, n] - x0[j, n])
        K1 = f + Fbrown
        Fbrown = np.sqrt(2*Dhat/h)*(dW + S[i])
        # estimate for slope at interval end
        x1 = x0 + h*K1
        f = np.zeros(x0.shape)
        for j in range(1, N):
            for n in range(3):
                f[j, n] += -k_over_xi*(x1[j, n] - x1[j-1, n])
                f[j-1, n] += -k_over_xi*(x1[j-1, n] - x1[j, n])
        K2 = f + Fbrown
        x0 = x0 + h * (K1 + K2)/2
        if np.abs(t[i] - t_save[save_i]) < rtol*np.abs(t_save[save_i]):
            x[save_i] = x0
            save_i += 1
    return x


@njit
def jit_confined_srk1(N, L, b, D, Aex, rx, ry, rz, t, t_save):
    """
    Add an elliptical confinement.

    Energy is like cubed of distance
    outside of ellipsoid, pointing normally back in. times some factor Aex
    controlling its strength.
    """
    rtol = 1e-5
    # derived parameters
    L0 = L/(N-1)  # length per bead
    bhat = np.sqrt(L0*b)  # mean squared bond length of discrete gaussian chain
    Nhat = L/b  # number of Kuhn lengths in chain
    Dhat = D*N/Nhat  # diffusion coef of a discrete gaussian chain bead
    k_over_xi = 3*Dhat/bhat**2
    # initial position
    x0 = np.zeros((N, 3))
    # x0 = np.cumsum(x0, axis=0)
    for i in range(1, N):
        # 1/sqrt(3) since generating per-coordinate
        x0[i] = x0[i-1] + bhat/np.sqrt(3)*np.random.randn(3)
        while x0[i, 0]**2/rx**2 + x0[i, 1]**2/ry**2 + x0[i, 2]**2/rz**2 > 1:
            x0[i] = x0[i-1] + bhat/np.sqrt(3)*np.random.randn(3)
    if t_save is None:
        t_save = t
    x = np.zeros(t_save.shape + x0.shape)
    dts = np.diff(t)
    # -1 or 1, p=1/2
    S = 2*(np.random.rand(len(t)) < 0.5) - 1
    save_i = 0
    if 0 == t_save[save_i]:
        x[0] = x0
        save_i += 1
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new functin later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = dts[i-1]
        dW = np.random.randn(*x0.shape)
        # D = sigma^2/2 ==> sigma = np.sqrt(2*D)
        Fbrown = np.sqrt(2*Dhat/h)*(dW - S[i])
        # estimate for slope at interval start
        f = np.zeros(x0.shape)
        j = 0
        conf = x0[j, 0]**2/rx**2 + x0[j, 1]**2/ry**2 + x0[j, 2]**2/rz**2
        if conf > 1:
            conf_u = np.array([
                -x0[j, 0]/rx**2, -x0[j, 1]/ry**2, -x0[j, 2]/rz**2
            ])
            conf_u = conf_u/np.linalg.norm(conf_u)
            f[j] += Aex*conf_u*np.power(np.sqrt(conf) - 1, 3)
        for j in range(1, N):
            conf = x0[j, 0]**2/rx**2 + x0[j, 1]**2/ry**2 + x0[j, 2]**2/rz**2
            if conf > 1:
                conf_u = np.array([
                    -x0[j, 0]/rx**2, -x0[j, 1]/ry**2, -x0[j, 2]/rz**2
                ])
                conf_u = conf_u/np.linalg.norm(conf_u)
                f[j] += Aex*conf_u*np.power(np.sqrt(conf) - 1, 3)
            for n in range(3):
                f[j, n] += -k_over_xi*(x0[j, n] - x0[j-1, n])
                f[j-1, n] += -k_over_xi*(x0[j-1, n] - x0[j, n])
        K1 = f + Fbrown
        Fbrown = np.sqrt(2*Dhat/h)*(dW + S[i])
        # estimate for slope at interval end
        x1 = x0 + h*K1
        f = np.zeros(x1.shape)
        j = 0
        conf = x1[j, 0]**2/rx**2 + x1[j, 1]**2/ry**2 + x1[j, 2]**2/rz**2
        if conf > 1:
            conf_u = np.array([
                -x1[j, 0]/rx**2, -x1[j, 1]/ry**2, -x1[j, 2]/rz**2
            ])
            conf_u = conf_u/np.linalg.norm(conf_u)
            f[j] += Aex*conf_u*np.power(np.sqrt(conf) - 1, 3)
        for j in range(1, N):
            conf = x1[j, 0]**2/rx**2 + x1[j, 1]**2/ry**2 + x1[j, 2]**2/rz**2
            if conf > 1:
                conf_u = np.array([
                    -x1[j, 0]/rx**2, -x1[j, 1]/ry**2, -x1[j, 2]/rz**2
                ])
                conf_u = conf_u/np.linalg.norm(conf_u)
                f[j] += Aex*conf_u*np.power(np.sqrt(conf) - 1, 3)
            for n in range(3):
                f[j, n] += -k_over_xi*(x1[j, n] - x1[j-1, n])
                f[j-1, n] += -k_over_xi*(x1[j-1, n] - x1[j, n])
        K2 = f + Fbrown
        x0 = x0 + h * (K1 + K2)/2
        if np.abs(t[i] - t_save[save_i]) < rtol*np.abs(t_save[save_i]):
            x[save_i] = x0
            save_i += 1
    return x


@njit
def jit_confinement_clean(N, L, b, D, Aex, rx, ry, rz, t, t_save):
    """
    Unfortunately, by "cleaning up" this function, we also make it 2x slower.

    #worthit
    """
    f_c = f_conf_ellipse
    f_e = f_elas_linear_rouse
    rtol = 1e-5
    # derived parameters
    L0 = L/(N-1)  # length per bead
    bhat = np.sqrt(L0*b)  # mean squared bond length of discrete gaussian chain
    Nhat = L/b  # number of Kuhn lengths in chain
    Dhat = D*N/Nhat  # diffusion coef of a discrete gaussian chain bead
    k_over_xi = 3*Dhat/bhat**2
    # initial position
    x0 = init_linear_rouse_conf(N, bhat, rx, ry, rz)
    # pre-alloc output
    if t_save is None:
        t_save = t
    x = np.zeros(t_save.shape + x0.shape)
    # setup for saving only requested time points
    save_i = 0
    if 0 == t_save[save_i]:
        x[0] = x0
        save_i += 1
    # at each step i, we use data (x,t)[i-1] to create (x,t)[i]
    # in order to make it easy to pull into a new function later, we'll call
    # t[i-1] "t0", old x (x[i-1]) "x0", and t[i]-t[i-1] "h".
    for i in range(1, len(t)):
        h = t[i] - t[i-1]
        dW = np.random.randn(*x0.shape)
        # -1 or 1, p=1/2
        S = 2*(np.random.rand() < 0.5) - 1
        # D = sigma^2/2 ==> sigma = np.sqrt(2*D)
        Fbrown = np.sqrt(2*Dhat/h)*(dW - S)
        # estimate for slope at interval start
        K1 = f_c(x0, Aex, rx, ry, rz) + f_e(x0, k_over_xi) + Fbrown
        Fbrown = np.sqrt(2*Dhat/h)*(dW + S)
        x1 = x0 + h*K1
        # estimate for slope at interval end
        K2 = f_c(x1, Aex, rx, ry, rz) + f_e(x1, k_over_xi) + Fbrown
        # average the slope estimates
        x0 = x0 + h * (K1 + K2)/2
        if np.abs(t[i] - t_save[save_i]) < rtol*np.abs(t_save[save_i]):
            x[save_i] = x0
            save_i += 1
    return x



"""Monte Carlo of inserting stuff inside of a sphere."""
from enum import Enum
from numba import jit
import numpy as np
import math
from .. import plot as wplot
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sns

SMALL_NUM = 10e-8

# @jit(nopython=True)
def is_inside(x, xr, r):
    """is the sphere of radius xr totally inside the confining sphere of radius
    r centered at the origin?
    ||x|| + xr < r ?"""
    return x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < (r - xr)*(r - xr)

# @jit(nopython=True)
def norm3_squared(x):
    return x[0]*x[0] + x[1]*x[1] + x[2]*x[2]

# @jit(nopython=True)
def weeks_chandler_anderson(r, sigma, epsilon=1):
    """WCA Interaction Potential (a rescaled "repulsive part of lennard jones"
    potential) with epsilon defaulting to 1, so that you can fix beta*epsilon
    elsewhere in the simulation."""
    # r = x2 - x1
    r2 = norm3_squared(r)
    if r2 < SMALL_NUM:
        return np.inf
    sigma2 = sigma*sigma
    if math.pow(2, 1/3)*sigma2 < r2:
        return 0
    sigr = sigma2/r2
    sigr = sigr*sigr*sigr
    return 4*epsilon*(sigr*(sigr - 1) + 1/4)

# # turns out this can be made into a clever weeks_chandler_anderson call
# @jit(nopython=True)
# def confined_in_sphere(x, r_x, r_conf, epsilon=1):
#     """E_confinement = E_WCA(r_conf - ||x||)."""
#     r = r_conf - np.linalg.norm(x)

def num_spheres_from_density(target_density, sphere_radii, confinement_radius):
    # num_spheres*sphere_volume = target_density*confinement_volume
    sphere_volume = volume_of_sphere(sphere_radii)
    confinement_volume = volume_of_sphere(confinement_radius)
    num_spheres = math.floor(target_density*confinement_volume/sphere_volume)
    return num_spheres

def initial_locations(num_spheres, sphere_radii, confinement_radius):
    """Get uniformly random initial locations for Metropolis Algorithms to
    start at."""
    num_successful = 0
    sphere_centers = np.zeros((num_spheres, 3))
    while num_successful < num_spheres:
        new_pos = 2*confinement_radius*(np.random.random_sample((3,)) - 1/2)
        if is_inside(new_pos, sphere_radii, confinement_radius):
            sphere_centers[num_successful, :] = new_pos
            num_successful += 1
    return sphere_centers

# @jit(nopython=True)
def volume_of_sphere(radius):
    return (4/3)*math.pi*(radius**3)

def single_wca_energy(sphere_pos, sphere_idx, sphere_centers, num_spheres, sphere_radii,
                      confinement_radius=1, energy_outside=np.inf):
    """Return contribution to energy of a particle at position sphere_pos if it
    replaced the particle with index sphere_idx in the list of particles
    sphere_centers corresponding to spheres of radius sphere_radii inside of a
    confining sphere of radius confinement_radius=1 centered at the origin.
    Distance from confinement has an energetic contribution equal to being that
    same distance from another sphere.

    Use energy_outside=np.inf to set the energy associated with being totally
    outside the confinement. Set energy_outside to None to honor the CWA
    potential symmetrically outside the sphere."""
    energy = 0
    dist_to_bdry = confinement_radius - np.linalg.norm(sphere_pos)
    if dist_to_bdry < 0 and energy_outside is not None:
        energy += energy_outside
    else:
        # vector with correct magnitude to pass to weeks_chandler_anderson
        vec_to_bdry = np.array([dist_to_bdry, 0, 0])
        energy += weeks_chandler_anderson(vec_to_bdry, 2*sphere_radii)
    for sj in range(num_spheres):
        if sj == sphere_idx:
            continue
        energy += weeks_chandler_anderson(sphere_pos - sphere_centers[sj,:], 2*sphere_radii)
    return energy

# @jit(nopython=True)
def total_wca_energy(sphere_centers, num_spheres, sphere_radii,
                     confinement_radius=1, energy_outside=np.inf):
    """Return total energy of particles in sphere_centers corresponding to
    spheres of radius sphere_radii inside of a confining sphere of radius
    confinement_radius=1 centered at the origin. Distance from confinement has an energetic
    contribution equal to being that same distance from another sphere.

    Use energy_outside=np.inf to set the energy associated with being totally
    outside the confinement. Set energy_outside to None to honor the CWA
    potential symmetrically outside the sphere."""
    energy = 0
    for si in range(num_spheres):
        # confinement energy
        dist_to_bdry = confinement_radius - np.linalg.norm(sphere_centers[si,:])
        if dist_to_bdry < 0 and energy_outside is not None:
            energy += energy_outside
        else:
            # vector with correct magnitude to pass to weeks_chandler_anderson
            vec_to_bdry = np.array([dist_to_bdry, 0, 0])
            energy += weeks_chandler_anderson(vec_to_bdry, 2*sphere_radii)
        # interaction energy
        for sj in range(si):
            energy += weeks_chandler_anderson(sphere_centers[si,:] - sphere_centers[sj,:], 2*sphere_radii)
    return energy

def norm_step_mc(num_steps, sphere_centers, num_spheres, sphere_radii, confinement_radius=1,
                 step_size=None, beta_epsilon=0.665, energy_outside=np.inf):
    """Peform num_steps monte carlo steps on the set of spheres with
    sphere_centers, sphere_radii in a confinement centered at the origin of
    radius confinement_radius. At each step, move one bead by a gaussian amount
    with std dev step_size(default->sphere_radii). The beads are assumed to
    have a weeks_chandler_anderson potential between them with sigma=sphere_radii, and
    epsilon and the temperature are determined by beta_epsilon. The confinement
    sphere is also weeks_chandler_anderson."""
    # default step size to be on average the size of the bead
    step_size = sphere_radii if step_size is None else step_size
    tot_energy_change = 0
    for i in range(num_steps):
        si = np.random.randint(num_spheres)
        # new positions energy calculation, and exit early if possible
        new_pos = sphere_centers[si,:] + step_size*np.random.standard_normal((3,))
        new_dist_to_bdry = confinement_radius - np.linalg.norm(new_pos)
        if new_dist_to_bdry < 0: # then you're outisde the bdry
            continue
        new_potential = single_wca_energy(new_pos, si, sphere_centers, num_spheres, sphere_radii, confinement_radius, energy_outside)
        old_potential = single_wca_energy(sphere_centers[si,:], si, sphere_centers, num_spheres, sphere_radii, confinement_radius, energy_outside)
        pot_diff = new_potential - old_potential
        # MH acceptance rule, most short-circuitable form
        if pot_diff > 0 or np.log(np.random.rand()) >= -beta_epsilon*pot_diff:
            continue
        # if we get here, the monte carlo step is accepted.
        sphere_centers[si,:] = new_pos
        tot_energy_change += pot_diff
    return sphere_centers, tot_energy_change

def sphere_dispersal_mc(num_steps, target_density, sphere_radii, confinement_radius=1,
                        steps_per_check=1000, step_size=None,
                        beta_epsilon=0.665, initial_centers=None):
    """Perform MCMC for num_steps after uniform position initialization of
    spheres with effective hard repulsive size sphere_radii (accomplished by
    weeks_chandler_anderson potential with barker-henderson mean collision
    diameter set to equal sigma) at target_density inside of a sphere with size
    confinement_radius.

    For now, beta_epsilon fixed to 0.665, per what Tom did in his thesis."""
    num_spheres = num_spheres_from_density(target_density, sphere_radii, confinement_radius)
    if initial_centers is None:
        sphere_centers = initial_locations(num_spheres, sphere_radii, confinement_radius)
    else:
        sphere_centers = initial_centers
    # break run into shorter sprints of 1000 steps, report energy change after
    # each 1000 steps
    num_checks = math.floor(num_steps/steps_per_check)
    energy = total_wca_energy(sphere_centers, num_spheres, sphere_radii, confinement_radius)
    for i in range(num_checks):
        # run MC, reports energy change
        sphere_centers, d_energy = norm_step_mc(num_steps=steps_per_check,
                sphere_centers=sphere_centers, num_spheres=num_spheres,
                sphere_radii=sphere_radii, confinement_radius=confinement_radius,
                step_size=step_size, beta_epsilon=beta_epsilon)
        print(energy + d_energy)
        energy = total_wca_energy(sphere_centers, num_spheres, sphere_radii, confinement_radius)
        print(energy)
    return sphere_centers

def mc_minimize_energy(target_density, sphere_radii, confinement_radius=1,
                  steps_per_check=10000, step_size=None, beta_epsilon=0.665,
                  initial_centers=None):
    """Perform MCMC after uniform position initialization of
    spheres with effective hard repulsive size sphere_radii (accomplished by
    weeks_chandler_anderson potential with barker-henderson mean collision
    diameter set to equal sigma) at target_density inside of a sphere with size
    confinement_radius. Rerun in batches of steps_per_check MC steps until the
    energies converge. To check if energy has converged, we do the lazy thing:
    wait till it increases for the first time then just run one more MC batch
    after that.

    For now, beta_epsilon fixed to 0.665, per what Tom did in his thesis."""
    num_spheres = num_spheres_from_density(target_density, sphere_radii, confinement_radius)
    if initial_centers is None:
        sphere_centers = initial_locations(num_spheres, sphere_radii, confinement_radius)
    else:
        sphere_centers = initial_centers
    energy = total_wca_energy(sphere_centers, num_spheres, sphere_radii, confinement_radius)
    i = 0
    while True:
        print("Total Energy after batch {i}: {energy: =10.8g}".format(i=i, energy=energy))
        sphere_centers, d_energy = norm_step_mc(num_steps=steps_per_check,
                sphere_centers=sphere_centers, num_spheres=num_spheres,
                sphere_radii=sphere_radii, confinement_radius=confinement_radius,
                step_size=step_size, beta_epsilon=beta_epsilon)
        i += 1
        energy += d_energy
        if d_energy > 0:
            print("Total Energy after batch {i}: {energy: =10.8g}".format(i=i, energy=energy))
            sphere_centers, d_energy = norm_step_mc(num_steps=steps_per_check,
                    sphere_centers=sphere_centers, num_spheres=num_spheres,
                    sphere_radii=sphere_radii, confinement_radius=confinement_radius,
                    step_size=step_size, beta_epsilon=beta_epsilon)
            i += 1
            energy += d_energy
            print("Total Energy after batch {i}: {energy: =10.8g}".format(i=i, energy=energy))
            break
    return sphere_centers

def plot_spheres(sphere_centers, radii, **kwargs):
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    palette = sns.color_palette('hls', 12)
    for i,c in enumerate(sphere_centers):
        color = palette[i % len(palette)]
        wplot.draw_sphere(c, radii, colors=color, axes=ax, **kwargs)
    return ax



"""
Routines for setting the initial locations of the beads for BD.

Functions here are largely used in order to make sure that the initial
configuration that is used to perform BD is valid.

.. note::

    Writing a correct ``init_*`` function can be arbitrarily difficult due to
    compounding factors such as confinement, bead connectivity, inter-polymer
    interactions, and so forth. If your system is complicated enough, consider
    using an existing function here and simply adding your energies to the
    Monte-Carlo code in order to generate approximately equilibrium initial
    configurations using MC.

This file also contains routines to directly initialize some simple types of
chains extremely near (or at) equilibrium to reduce the number of
initialization BD or MC steps required to equilibrate the polymer.
Alternatively, some of these chains are as far as possible from equilibrium
(such as `init_straight_x`) in order to ensure that e.g. the end-to-end
distance reaching its equilibrium distribution is representative of the time it
takes for the chain itself to reach equilibrium.


.. note::

    If you know the analog of the "Rouse" time for your system (i.e. the
    time-scale of the largest mode), this is typically an easier and  more
    principled way to determine how long to run initialization BD than doing a
    huge number of simulations and trying to measure how long it takes for
    something like the end-to-end distribution to equilibrate.
"""
import numpy as np
from numba import jit


def init_straight_x(N, L):
    """Return straight line from 0 along positive x-axis."""
    x0 = np.zeros((N, 3))
    x0[:, 0] = np.linspace(0, L, N)
    return x0


def init_circle(N, L):
    """Return a circle with circumference L in the x-y plane."""
    # want circumference L
    R = L/(2*np.pi)
    r = np.zeros((N, 3))
    r[:, 0] = R*np.cos(np.linspace(0, 2*np.pi, N))
    r[:, 1] = R*np.sin(np.linspace(0, 2*np.pi, N))
    u = np.zeros((N, 3))
    u[:, 0] = -np.sin(np.linspace(0, 2*np.pi, N))
    u[:, 1] = np.cos(np.linspace(0, 2*np.pi, N))
    return r, u


@jit(nopython=True)
def init_linear_rouse_conf(N, bhat, rx, ry, rz):
    """Ad-hoc method for Rouse polymers in elliptical confinement.

    Will not necessarily be exactly at equilibrium, but pretty close.

    Parameters
    ----------
    bhat : float
        b*np.sqrt(Lb), the std dev of the distance between beads
    rx, ry, rz : float
        Principle axes of confinement ellipse

    Returns
    -------
    (N, 3) array of float
        The initial locations of the beads.
    """
    x0 = np.zeros((N, 3))
    # x0 = np.cumsum(x0, axis=0)
    for i in range(1, N):
        # 1/sqrt(3) since generating per-coordinate
        x0[i] = x0[i-1] + bhat/np.sqrt(3)*np.random.randn(3)
        while x0[i, 0]**2/rx**2 + x0[i, 1]**2/ry**2 + x0[i, 2]**2/rz**2 > 1:
            x0[i] = x0[i-1] + bhat/np.sqrt(3)*np.random.randn(3)
    return x0


@jit(nopython=True)
def init_homolog_rouse_conf(N, N_tot, loop_list, bhat, rx, ry, rz):
    """VERY ad-hoc, probably not near equilibrium, or even in confinement."""
    # first initialize the first chain
    x0 = np.zeros((N_tot, 3))
    x0[:N, :] = init_linear_rouse_conf(N, bhat, rx, ry, rz)
    # now initialize the loops
    num_loops, _ = loop_list.shape
    for j in range(1, num_loops-1):
        k1l, k1r, k2l, k2r = loop_list[j]
        # need to initialize beads [k2l,..,k2r] so that they form a Brownian
        # bridge between the (already determined) positions of k1l and k1r
        # first, build a regular brownian walk starting at k1l
        num_beads = k1r - k1l - 1
        for i in range(num_beads):
            i1 = k2l + i - 1 if i > 0 else k1l
            i2 = k2l + i
            x0[i2] = x0[i1] + bhat/np.sqrt(3)*np.random.randn(3)
            while 1 < x0[i2, 0]**2/rx**2 + x0[i2, 1]**2/ry**2 \
                    + x0[i2, 2]**2/rz**2:
                x0[i2] = x0[i1] + bhat/np.sqrt(3)*np.random.randn(3)
        # then subtract off the correct amount from each step to make the
        # brownian bridge. This guy is no longer guaranteed to be in the
        # confinement....but will typically not be catastrophically far outside
        bridge = x0[k1r] - x0[k1l]  # target end-to-end vector
        # center bridge temporarily about zero
        for jj in range(k2l, k2r+1):
            x0[jj] -= x0[k1l]
        # jit >:( x0[k2l:k2r+1] = x0[k2l:k2r+1] - x0[k1l][None,:]
        # now fix end-to-end vector
        for i in range(num_beads):
            # fraction of the way along the bridge
            f = (i + 1)/(num_beads + 1)
            x0[k2l+i] = x0[k2l+i]*(1 - f) + f*bridge
        # re-center bridge around x0[k1l]
        for jj in range(k2l, k2r+1):
            x0[jj] += x0[k1l]
        # x0[k2l:k2r+1] = x0[k2l:k2r+1] + x0[k1l][None,:]
    # finally, initialize the free ends
    if num_loops == 1:
        x0[N:, :] = init_linear_rouse_conf(N, bhat, rx, ry, rz)
        return x0
    # "left" free end must be built backwards from first connection point
    k1l, k1r, k2l, k2r = loop_list[0]
    num_beads = k1r - k1l - 1  # or k2r - k2l + 1
    for i in range(num_beads):
        i1 = k2r - i + 1 if i > 0 else k1r
        i2 = k2r - i
        x0[i2] = x0[i1] + bhat/np.sqrt(3)*np.random.randn(3)
        while x0[i2, 0]**2/rx**2 + x0[i2, 1]**2/ry**2 + x0[i2, 2]**2/rz**2 > 1:
            x0[i2] = x0[i1] + bhat/np.sqrt(3)*np.random.randn(3)
    # right free end can be built as normal with no bridge stuff
    k1l, k1r, k2l, k2r = loop_list[-1]
    num_beads = k1r - k1l - 1  # or k2r - k2l + 1
    for i in range(num_beads):
        i1 = k2l + i - 1 if i > 0 else k1l
        i2 = k2l + i
        x0[i2] = x0[i1] + bhat/np.sqrt(3)*np.random.randn(3)
        while x0[i2, 0]**2/rx**2 + x0[i2, 1]**2/ry**2 + x0[i2, 2]**2/rz**2 > 1:
            x0[i2] = x0[i1] + bhat/np.sqrt(3)*np.random.randn(3)
    return x0
